//******************************************************************
//*******************************************************************
//*** Client
//*******************************************************************
//*******************************************************************

//-------------------------------------------------------------------
//  Required definitions
//-------------------------------------------------------------------
#define UNICODE
#define _WIN32_DCOM

//-------------------------------------------------------------------
//  includes
//-------------------------------------------------------------------
#include <assert.h>
#include <stdio.h>
#include <windows.h>
#include <initguid.h>
#include "..\idl\details.h"

//-------------------------------------------------------------------
//  includes
//-------------------------------------------------------------------
CRITICAL_SECTION g_cs;
FILE *g_pFile = NULL;

//*******************************************************************
//*******************************************************************
//***  UTILITY FUNCTIONS
//*******************************************************************
//*******************************************************************
void DisplayStatus(wchar_t *pwszMsg, HRESULT hr)
{

   if (hr == S_OK)
   {
      wprintf(TEXT("%s\n"), pwszMsg);
      return;
   }

   if (HRESULT_FACILITY(hr) == FACILITY_WINDOWS)
      hr = HRESULT_CODE(hr);

   wchar_t *pwszStatus;
   FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
      FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      hr,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      (LPWSTR)&pwszStatus,
      0,
      NULL );

   wprintf(TEXT("%s: %s (ECode: %lx)\n"), pwszMsg, pwszStatus, hr);

   LocalFree(pwszStatus);
}

//-------------------------------------------------------------------
// Performance testing
//-------------------------------------------------------------------
#define QUERYPERFORMANCE(code) \
{   double elapseTime = 0; \
   LARGE_INTEGER   StartTime, StopTime, Frequency; \
   QueryPerformanceCounter(&StartTime); \
   ##code \
   QueryPerformanceCounter(&StopTime); \
   QueryPerformanceFrequency(&Frequency); \
   elapseTime = (double)(( StopTime.LowPart - StartTime.LowPart)/(double)Frequency.LowPart);\
   fwprintf(g_pFile, TEXT("%f\t"), elapseTime); \
}

//-------------------------------------------------------------------
// Performance testing
//-------------------------------------------------------------------
void PerformanceTesting(DWORD dwClsContext)
{
   IDetails *pDetails = 0 ;
   HRESULT hr = S_OK;

   wchar_t Type[MAX_PATH];
   switch (dwClsContext) {
   case CLSCTX_INPROC_SERVER: wcscpy(Type, TEXT("InProc")); break;
   case CLSCTX_LOCAL_SERVER: wcscpy(Type, TEXT("Local")); break;
   case CLSCTX_REMOTE_SERVER: wcscpy(Type, TEXT("Remote"));break;
   default: wcscpy(Type, TEXT("Unknown")); break;
   }

   fwprintf(g_pFile, TEXT("%s\t"), Type); 
   QUERYPERFORMANCE(
      hr = CoCreateInstance(CLSID_Details, 0, dwClsContext,
            IID_IDetails, reinterpret_cast<void**>(&pDetails));
   );

   if (SUCCEEDED(hr)) {
      QUERYPERFORMANCE(pDetails->VoidCall();); 
      QUERYPERFORMANCE(pDetails->VoidCall();); 
      QUERYPERFORMANCE(long ArrayLong[10]; pDetails->ArrayCall(10, ArrayLong););  
      QUERYPERFORMANCE(for (int i=0; i<10; i++) {pDetails->LongCall(1);}); 
      pDetails->Release();
   } else {
      DisplayStatus(TEXT("Failed CoCreateInstance"), hr);
   }
   fwprintf(g_pFile, TEXT("\n")); 
}

//-------------------------------------------------------------------
// The second STA thread.  This thread enters and execute within its
// own STA apartment.  An STA can hold only one thread
//-------------------------------------------------------------------
DWORD WINAPI STAThread(LPVOID)
{
   DisplayStatus(TEXT("STAThread"), S_OK);
   // A thread must enter an apartment before it can use COM
   HRESULT hr = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED);
   assert(SUCCEEDED(hr));

   // performance testing
   EnterCriticalSection(&g_cs);
   fwprintf(g_pFile, TEXT("Secondary STA Thread\n"));
   fwprintf(g_pFile, TEXT("\tActivation\tFirstCall\tVoidCall\tArrayCall\tLongCall\n"));
   PerformanceTesting(CLSCTX_INPROC_SERVER);
   PerformanceTesting(CLSCTX_LOCAL_SERVER);
   PerformanceTesting(CLSCTX_REMOTE_SERVER);
   fflush(g_pFile);
   LeaveCriticalSection(&g_cs);

   // Marshaling demonstration

   CoUninitialize();
   return 0;
}

//-------------------------------------------------------------------
// MTAThread1 is a thread that enters into the ONE-AND-ONLY MTA
// within this process.  This thread executes only within this
// MTA aparment.  And MTA can host many thread.
//-------------------------------------------------------------------
DWORD WINAPI MTAThread1(LPVOID)
{
   DisplayStatus(TEXT("MTAThread1"), S_OK);
   // A thread must enter an apartment before it can use COM
   HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
   assert(SUCCEEDED(hr));

   // performance testing
   EnterCriticalSection(&g_cs);
   fwprintf(g_pFile, TEXT("MTAThread1\n"));
   fwprintf(g_pFile, TEXT("\tActivation\tFirstCall\tVoidCall\tArrayCall\tLongCall\n"));
   PerformanceTesting(CLSCTX_INPROC_SERVER);
   PerformanceTesting(CLSCTX_LOCAL_SERVER);
   PerformanceTesting(CLSCTX_REMOTE_SERVER);
   fflush(g_pFile);
   LeaveCriticalSection(&g_cs);

   // Marshaling demonstration

   CoUninitialize();
   return 0;
}

//-------------------------------------------------------------------
// The second STA thread.  This thread enters and execute within its
// own STA apartment.
//-------------------------------------------------------------------
DWORD WINAPI MTAThread2(LPVOID)
{
   DisplayStatus(TEXT("MTAThread2"), S_OK);
   // A thread must enter an apartment before it can use COM
   HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
   assert(SUCCEEDED(hr));

   // performance testing
   EnterCriticalSection(&g_cs);
   fwprintf(g_pFile, TEXT("MTAThread2\n"));
   fwprintf(g_pFile, TEXT("\tActivation\tFirstCall\tVoidCall\tArrayCall\tLongCall\n"));
   PerformanceTesting(CLSCTX_INPROC_SERVER);
   PerformanceTesting(CLSCTX_LOCAL_SERVER);
   PerformanceTesting(CLSCTX_REMOTE_SERVER);
   fflush(g_pFile);
   LeaveCriticalSection(&g_cs);

   // Marshaling demonstration

   CoUninitialize();
   return 0;
}


//*******************************************************************
//*******************************************************************
//*** Main
//*******************************************************************
//*******************************************************************
void main(int argc, char **argv)
{
   g_pFile = _wfopen(TEXT("perform.log"), TEXT("a"));
   fwprintf(g_pFile, TEXT("\n***Started New Log***\n"));

   //-----------------------------------------
   // This is the main thread
   //-----------------------------------------
   DisplayStatus(TEXT("Client: Main STA Started"), S_OK);
   InitializeCriticalSection(&g_cs);
   HRESULT hr = S_OK;

   //-----------------------------------------
   // Init COM - this is an STA
   //-----------------------------------------
   hr = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED);
   assert(SUCCEEDED(hr));

   //-----------------------------------------
   // Spawn to other threads
   //-----------------------------------------
   DWORD dwTidSTAThread = 0;
   HANDLE m_hSTAThread = CreateThread(0, 0, STAThread, 0, 
            CREATE_SUSPENDED, &dwTidSTAThread);

   DWORD dwTidMTAThread1 = 0;
   HANDLE m_hMTAThread1 = CreateThread(0, 0, MTAThread1, 0, 
            CREATE_SUSPENDED, &dwTidMTAThread1);

   DWORD dwTidMTAThread2 = 0;
   HANDLE m_hMTAThread2 = CreateThread(0, 0, MTAThread2, 0, 
            CREATE_SUSPENDED, &dwTidMTAThread2);

   if (m_hSTAThread && m_hMTAThread1 && m_hMTAThread2) {
      ResumeThread(m_hSTAThread);
      ResumeThread(m_hMTAThread1);
      ResumeThread(m_hMTAThread2);
   } else {
      assert(false);
   }

   DisplayStatus(TEXT("\n***Ctrl-C to end...\n"), S_OK);

   //-----------------------------------------
   // STAs must have a message pump that 
   // dispatches windows messages
   //-----------------------------------------
   MSG m;
   while (GetMessage(&m, NULL, 0, 0)) { DispatchMessage (&m); }

   //-----------------------------------------
   // Main thread goes away - child threads 
   // will die off uncleanly
   //-----------------------------------------

   //-----------------------------------------
   // Shutdown COM
   //-----------------------------------------
   CoUninitialize();
   DisplayStatus(TEXT("Client shutting down..."), S_OK);
   DeleteCriticalSection(&g_cs);
   fclose(g_pFile);
}
